"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const match_1 = require("./translate/match");
const printf_1 = require("./translate/printf");
var Fallback;
(function (Fallback) {
    Fallback[Fallback["throw"] = 0] = "throw";
    Fallback[Fallback["id"] = 1] = "id";
    Fallback[Fallback["skip"] = 2] = "skip";
})(Fallback = exports.Fallback || (exports.Fallback = {}));
const initial_options = {
    datas: undefined,
    fallback: Fallback.throw,
    start_file: 'stat_descriptions'
};
const formatStats = Object.assign((stats, options = {}) => {
    const { datas, fallback, start_file } = Object.assign({}, formatStats.options, options);
    if (datas === undefined) {
        throw new Error('locale datas not provided. Set it either via passed option or #configure');
    }
    // translated lines
    const lines = [];
    // array of stat_ids for which hash lookup failed
    const untranslated = new Map(stats.map((stat) => [stat.id, stat]));
    let description_file = datas[start_file];
    while (description_file !== undefined) {
        const data = description_file.data;
        lines.push(...formatWithFinder(untranslated, ({ id }) => data[id]));
        lines.push(...formatWithFinder(untranslated, ({ id }) => findDescription(id, data)));
        description_file = description_file.meta.include
            ? datas[description_file.meta.include]
            : undefined;
    }
    lines.push(...formatWithFallback(untranslated, fallback));
    return lines;
}, {
    options: initial_options,
    configure: (options) => {
        formatStats.options = Object.assign({}, formatStats.options, options);
    }
});
exports.default = formatStats;
/**
 * O(n) lookup if hash lookup fails
 *
 * @param stat_id
 * @param locale_data
 */
function findDescription(stat_id, locale_data) {
    return Object.values(locale_data).find(({ stats }) => stats.includes(stat_id));
}
const NO_DESCRIPTION = 'NO_DESCRIPTION';
// stats will get mutated
function formatWithFinder(stats, find) {
    const lines = [];
    for (const [stat_id, stat] of stats) {
        const description = find(stat);
        if (description !== undefined) {
            const translation = translate(description, stats);
            if (translation === undefined) {
                throw new Error(`matching translation not found for '${stat.id}'`);
            }
            else {
                // mark as translated
                description.stats.forEach(translated_id => stats.delete(translated_id));
                if (translation === NO_DESCRIPTION) {
                    lines.push(`${stat_id} (hidden)`);
                }
                else {
                    lines.push(translation);
                }
            }
        }
    }
    return lines;
}
function translate(description, provided) {
    const { stats, no_description, translations } = description;
    if (no_description === true) {
        return NO_DESCRIPTION;
    }
    // intersect the required stat_ids from the desc with the provided
    const required_stats = stats
        .map(stat_id => {
        const stat = provided.get(stat_id);
        // default the value to 0
        if (stat === undefined) {
            return {
                id: stat_id,
                value: 0
            };
        }
        else {
            return stat;
        }
    })
        .filter((stat) => stat !== null);
    const translation = matchingTranslation(translations, required_stats);
    if (translation === undefined) {
        return undefined;
    }
    else {
        return printf_1.default(translation.text, required_stats.map(({ value }) => value), translation.formatters);
    }
}
function matchingTranslation(translations, stats) {
    const args = stats.map(({ value }) => value);
    return translations.find(translation => {
        return match_1.matches(args, translation.matchers).every(match => match === match_1.Match.subset || match === match_1.Match.exact);
    });
}
function formatWithFallback(stats, fallback) {
    if (fallback === Fallback.throw) {
        if (stats.size > 0) {
            throw new Error('no descriptions found for ' + Array.from(stats.keys()).join(','));
        }
        else {
            return [];
        }
    }
    else if (fallback === Fallback.id) {
        return Array.from(stats.keys());
    }
    else if (fallback === Fallback.skip) {
        return [];
    }
    else if (typeof fallback === 'function') {
        return Array.from(stats.entries())
            .map(([id, stat]) => fallback(id, stat))
            .filter((line) => typeof line === 'string');
    }
    else {
        // should ts recognize that this is unreachable code? enums can prob
        // be extended at runtime an therfore somebody could mess with them
        throw new Error(`unrecognized fallback type '${fallback}'`);
    }
}
